import 'dart:convert';
import 'dart:io';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';

/// [BaseContainer]
abstract class _IBaseContainer {
  Widget _childBuild(BuildContext context);

  Widget _childTransformBuilder(Widget superCurrent);
}

///[Container]
abstract class BaseContainer extends StatelessWidget
    implements _IBaseContainer {
  final VoidCallback? onPressed;

  final VoidCallback? onLongPress;

  final bool? visible;

  final bool? isClipOval;

  final double? radius;

  final WidgetBuilder? errorBuild;

  final Function(BuildContext context, Widget current)? transformBuilder;

  BaseContainer({
    Key? key,
    this.alignment,
    this.padding,
    this.color,
    this.decoration,
    this.foregroundDecoration,
    double? width,
    double? height,
    BoxConstraints? constraints,
    this.margin,
    this.transform,
    this.transformAlignment,
    this.clipBehavior = Clip.none,
    this.onPressed,
    this.onLongPress,
    this.visible,
    this.isClipOval,
    this.radius,
    this.errorBuild,
    this.transformBuilder,
  })  : assert(margin == null || margin.isNonNegative),
        assert(padding == null || padding.isNonNegative),
        assert(decoration == null || decoration.debugAssertIsValid()),
        assert(constraints == null || constraints.debugAssertIsValid()),
        assert(clipBehavior != null),
        assert(decoration != null || clipBehavior == Clip.none),
        assert(
          color == null || decoration == null,
          'Cannot provide both a color and a decoration\n'
          'To provide both, use "decoration: BoxDecoration(color: color)".',
        ),
        constraints = (width != null || height != null)
            ? constraints?.tighten(width: width, height: height) ??
                BoxConstraints.tightFor(width: width, height: height)
            : constraints,
        super(key: key);

  late final Widget? _child;

  final AlignmentGeometry? alignment;

  final EdgeInsetsGeometry? padding;

  final Color? color;

  final Decoration? decoration;

  /// The decoration to paint in front of the [_child].
  final Decoration? foregroundDecoration;

  final BoxConstraints? constraints;

  /// Empty space to surround the [decoration] and [_child].
  final EdgeInsetsGeometry? margin;

  /// The transformation matrix to apply before painting the container.
  final Matrix4? transform;

  final AlignmentGeometry? transformAlignment;

  final Clip clipBehavior;

  EdgeInsetsGeometry? get _paddingIncludingDecoration {
    if (decoration == null || decoration!.padding == null) return padding;
    final EdgeInsetsGeometry? decorationPadding = decoration!.padding;
    if (padding == null) return decorationPadding;
    return padding!.add(decorationPadding!);
  }

  @override
  Widget build(BuildContext context) {
    _child = _childBuild(context);
    Widget? current = _child;

    if (_child == null && (constraints == null || !constraints!.isTight)) {
      current = LimitedBox(
        maxWidth: 0.0,
        maxHeight: 0.0,
        child: ConstrainedBox(constraints: const BoxConstraints.expand()),
      );
    }

    if (alignment != null)
      current = Align(alignment: alignment!, child: current);

    final EdgeInsetsGeometry? effectivePadding = _paddingIncludingDecoration;
    if (effectivePadding != null)
      current = Padding(padding: effectivePadding, child: current);

    if (color != null) current = ColoredBox(color: color!, child: current);

    if (clipBehavior != Clip.none) {
      assert(decoration != null);
      current = ClipPath(
        clipper: _DecorationClipper(
          textDirection: Directionality.maybeOf(context),
          decoration: decoration!,
        ),
        clipBehavior: clipBehavior,
        child: current,
      );
    }

    if (decoration != null)
      current = DecoratedBox(decoration: decoration!, child: current);

    if (foregroundDecoration != null) {
      current = DecoratedBox(
        decoration: foregroundDecoration!,
        position: DecorationPosition.foreground,
        child: current,
      );
    }

    if (constraints != null)
      current = ConstrainedBox(constraints: constraints!, child: current);

    current = _childTransformBuilder(current!);

    if (onPressed != null || onLongPress != null)
      current = GestureDetector(
          onTap: onPressed, onLongPress: onPressed, child: current);

    if (isClipOval == true) {
      current = ClipOval(
        child: current,
      );
    } else if (radius != null && radius! > 0) {
      current = ClipRRect(
        borderRadius: BorderRadius.circular(radius!),
        child: current,
      );
    }

    if (margin != null) current = Padding(padding: margin!, child: current);

    if (transform != null)
      current = Transform(
          transform: transform!, child: current, alignment: transformAlignment);

    if (transformBuilder != null)
      current = transformBuilder!.call(context, current);

    if (visible != null)
      current = Visibility(visible: visible!, child: current!);

    return current!;
  }

  @override
  Widget _childTransformBuilder(Widget superCurrent) => superCurrent;

  @override
  void debugFillProperties(DiagnosticPropertiesBuilder properties) {
    super.debugFillProperties(properties);
    properties.add(DiagnosticsProperty<AlignmentGeometry>(
        'alignment', alignment,
        showName: false, defaultValue: null));
    properties.add(DiagnosticsProperty<EdgeInsetsGeometry>('padding', padding,
        defaultValue: null));
    properties.add(DiagnosticsProperty<Clip>('clipBehavior', clipBehavior,
        defaultValue: Clip.none));
    if (color != null)
      properties.add(DiagnosticsProperty<Color>('bg', color));
    else
      properties.add(DiagnosticsProperty<Decoration>('bg', decoration,
          defaultValue: null));
    properties.add(DiagnosticsProperty<Decoration>('fg', foregroundDecoration,
        defaultValue: null));
    properties.add(DiagnosticsProperty<BoxConstraints>(
        'constraints', constraints,
        defaultValue: null));
    properties.add(DiagnosticsProperty<EdgeInsetsGeometry>('margin', margin,
        defaultValue: null));
    properties.add(ObjectFlagProperty<Matrix4>.has('transform', transform));
  }
}

/// A clipper that uses [Decoration.getClipPath] to clip.
class _DecorationClipper extends CustomClipper<Path> {
  _DecorationClipper({
    TextDirection? textDirection,
    required this.decoration,
  })  : assert(decoration != null),
        textDirection = textDirection ?? TextDirection.ltr;

  final TextDirection textDirection;
  final Decoration decoration;

  @override
  Path getClip(Size size) {
    return decoration.getClipPath(Offset.zero & size, textDirection);
  }

  @override
  bool shouldReclip(_DecorationClipper oldClipper) {
    return oldClipper.decoration != decoration ||
        oldClipper.textDirection != textDirection;
  }
}

///[TextWidget]
class TextWidget extends BaseContainer {
  final String text;

  final Color? backgroundColor;

  final InlineSpan? textSpan;

  final TextStyle? style;

  ///使用的支撑杆样式。撑杆样式定义撑杆，该撑杆设置最小的垂直布局度量
  final StrutStyle? strutStyle;

  ///Text()
  ///如何水平对齐
  final TextAlign? textAlign;

  final TextDirection? textDirection;

  final Locale? locale;

  final bool? softWrap;

  final TextOverflow overflow;

  final double? textScaleFactor;

  final int? maxLines;

  final String? semanticsLabel;

  final TextWidthBasis? textWidthBasis;

  final TextHeightBehavior? textHeightBehavior;

  ///
  final double? width;

  final double? height;

  final WidgetBuilder? errorBuild;

  TextWidget(this.text,
      {Key? key,
      this.width,
      this.height,
      this.errorBuild,
      this.textSpan,
      this.style,
      this.strutStyle,
      this.textAlign,
      this.textDirection,
      this.locale,
      this.softWrap,
      this.overflow = TextOverflow.ellipsis,
      this.textScaleFactor,
      this.maxLines,
      this.semanticsLabel,
      this.textWidthBasis,
      this.textHeightBehavior,
      this.onPressed,
      this.onLongPress,
      this.visible,
      this.isClipOval,
      this.radius,
      this.transformBuilder,
      this.alignment,
      this.padding,
      this.color,
      this.decoration,
      this.foregroundDecoration,
      this.constraints,
      this.margin,
      this.transform,
      this.transformAlignment,
      this.clipBehavior = Clip.none,
      this.backgroundColor})
      : super(
          key: key,
          width: width,
          height: height,
          errorBuild: errorBuild,
          onPressed: onPressed,
          onLongPress: onLongPress,
          visible: visible,
          isClipOval: isClipOval,
          radius: radius,
          transformBuilder: transformBuilder,
          alignment: alignment,
          padding: padding,
          color: color,
          decoration: decoration,
          foregroundDecoration: foregroundDecoration,
          constraints: constraints,
          margin: margin,
          transform: transform,
          transformAlignment: transformAlignment,
          clipBehavior: clipBehavior,
        );

  ///Base
  final VoidCallback? onPressed;

  final VoidCallback? onLongPress;

  final bool? visible;

  final bool? isClipOval;

  final double? radius;

  final Function(BuildContext context, Widget current)? transformBuilder;

  final AlignmentGeometry? alignment;

  final EdgeInsetsGeometry? padding;

  final Color? color;

  final Decoration? decoration;

  final Decoration? foregroundDecoration;

  final BoxConstraints? constraints;

  final EdgeInsetsGeometry? margin;

  final Matrix4? transform;

  final AlignmentGeometry? transformAlignment;

  final Clip clipBehavior;

  @override
  Widget _childBuild(BuildContext context) {
    return Text(
      (text == "null") ? '' : text,
      maxLines: maxLines,
      textAlign: textAlign,
      style: style,
      strutStyle: strutStyle,
      textDirection: textDirection,
      locale: locale,
      softWrap: softWrap,
      overflow: overflow,
      textScaleFactor: textScaleFactor,
      semanticsLabel: semanticsLabel,
      textWidthBasis: textWidthBasis,
      textHeightBehavior: textHeightBehavior,
    );
  }

  @override
  Widget _childTransformBuilder(Widget superCurrent) {
    if (backgroundColor != null)
      superCurrent = ColoredBox(color: backgroundColor!, child: superCurrent);
    return super._childTransformBuilder(superCurrent);
  }
}

///[ImageWidget]
typedef ImageWidgetBuilder = Widget? Function(ImageWidget);

const _DEFAULT_PACKAGE = "defaultPackage";

class ImageWidgetComponent {
  static ImageWidgetComponent? _instance;

  static ImageWidgetComponent get instance {
    return _instance ?? ImageWidgetComponent.init();
  }

  ImageWidgetBuilder? assetDebugBuilder;
  final ImageWidgetBuilder netWorkWidgetBuilder;
  final ImageWidgetBuilder svgWidgetBuilder;
  ImageWidgetBuilder? errorWidgetBuilder;
  final List<String> assetsNames;
  final String defaultPackage;

  ImageWidgetComponent._({
    this.assetDebugBuilder,
    required this.netWorkWidgetBuilder,
    required this.svgWidgetBuilder,
    this.errorWidgetBuilder,
    this.assetsNames = const ['assets/', 'image/', 'images/'],
    this.defaultPackage = _DEFAULT_PACKAGE,
  });

  static ImageWidgetBuilder _defaultBuilder = (iv) => null;

  factory ImageWidgetComponent.init(
      {ImageWidgetBuilder? assetDebugBuilder,
      ImageWidgetBuilder? netWorkWidgetBuilder,
      ImageWidgetBuilder? svgWidgetBuilder,
      ImageWidgetBuilder? errorBuilder,
      List<String> assetsNames = const ['assets/', 'image/', 'images/'],
      String defaultPackage = _DEFAULT_PACKAGE}) {
    if (_instance == null) {
      _instance = ImageWidgetComponent._(
        assetDebugBuilder: assetDebugBuilder,
        netWorkWidgetBuilder: netWorkWidgetBuilder ?? _defaultBuilder,
        svgWidgetBuilder: svgWidgetBuilder ?? _defaultBuilder,
        errorWidgetBuilder: errorBuilder,
        assetsNames: assetsNames,
        defaultPackage: defaultPackage,
      );
    }
    return _instance!;
  }

// ImageWidgetBuilder get svgWidgetBuilder {
//   return _svgWidgetBuilder;
// }
//
// ImageWidgetBuilder get netWorkWidgetBuilder {
//   return _netWorkWidgetBuilder;
// }
//
// ImageWidgetBuilder get assetDebugBuilder {
//   return _assetDebugBuilder;
// }
//
// ImageWidgetBuilder get errorWidgetBuilder {
//   return _errorWidgetBuilder;
// }
}

class ImageWidget extends BaseContainer {
  final String url;

  final String? errorUrl;

  final String? defUrl;

  final String? package;

  final BoxFit? fit;

  final double? width;

  final double? height;

  final WidgetBuilder? errorBuild;

  ImageWidget(this.url,
      {Key? key,
      this.errorUrl,
      this.defUrl,
      this.package,
      this.fit,
      this.width,
      this.height,
      this.errorBuild,
      this.onPressed,
      this.onLongPress,
      this.visible,
      this.isClipOval,
      this.radius,
      this.transformBuilder,
      this.alignment,
      this.padding,
      this.color,
      this.decoration,
      this.foregroundDecoration,
      this.constraints,
      this.margin,
      this.transform,
      this.transformAlignment,
      this.clipBehavior = Clip.none})
      : super(
          key: key,
          width: width,
          height: height,
          errorBuild: errorBuild,
          onPressed: onPressed,
          onLongPress: onLongPress,
          visible: visible,
          isClipOval: isClipOval,
          radius: radius,
          transformBuilder: transformBuilder,
          alignment: alignment,
          padding: padding,
          color: color,
          decoration: decoration,
          foregroundDecoration: foregroundDecoration,
          constraints: constraints,
          margin: margin,
          transform: transform,
          transformAlignment: transformAlignment,
          clipBehavior: clipBehavior,
        );

  ///Base
  final VoidCallback? onPressed;

  final VoidCallback? onLongPress;

  final bool? visible;

  final bool? isClipOval;

  final double? radius;

  final Function(BuildContext context, Widget current)? transformBuilder;

  final AlignmentGeometry? alignment;

  final EdgeInsetsGeometry? padding;

  final Color? color;

  final Decoration? decoration;

  final Decoration? foregroundDecoration;

  final BoxConstraints? constraints;

  final EdgeInsetsGeometry? margin;

  final Matrix4? transform;

  final AlignmentGeometry? transformAlignment;

  final Clip clipBehavior;

  @override
  Widget _childBuild(BuildContext context) {
    if (url.isEmpty || url == 'null') {
      return Container(
          width: width,
          height: height,
          color: color,
          child: _buildErrorWidget(context,
              msg: "url == null || url.isEmpty || url == 'null'"));
    }
    if (url.startsWith('https:') || url.startsWith('http:')) {
      return ImageWidgetComponent.instance.netWorkWidgetBuilder.call(this) ??
          Image.network(url, width: width, height: height, fit: fit,
              errorBuilder: (ctx, e, s) {
            return _buildErrorWidget(context, msg: "network error = $e");
          });
    } else if (ImageWidgetComponent.instance.assetsNames
        .any((element) => url.startsWith(element))) {
      /// assets
      Widget? result;

      assert(() {
        try {
          result = ImageWidgetComponent.instance.assetDebugBuilder?.call(this);
        } on Exception catch (e) {
          if (ImageWidgetComponent.instance.assetDebugBuilder != null) {
            result =
                _buildErrorWidget(context, msg: "assetDebugBuilder error = $e");
          }
        }
        return true;
      }());

      if (result != null) return result!;

      String? _package =
          package ?? ImageWidgetComponent.instance.defaultPackage;
      assert(_package.isNotEmpty, "未设置资源图片的bundle名称");
      if (_package == _DEFAULT_PACKAGE) _package = null;
      if (url.endsWith(".svg")) {
        assert(() {
          try {
            result = ImageWidgetComponent.instance.svgWidgetBuilder(this) ??
                _buildErrorWidget(context, msg: "svgWidgetBuilder == null");
          } catch (e) {
            result =
                _buildErrorWidget(context, msg: "svgWidgetBuilder error = $e");
          }
          return true;
        }());
        if (result != null) return result!;
        return ImageWidgetComponent.instance.svgWidgetBuilder(this) ??
            _buildErrorWidget(context, msg: "svgWidgetBuilder == null");
      }
      return Image.asset(url,
          fit: fit,
          package: _package,
          width: width,
          height: height,
          scale: 1.0);
    } else if (url.startsWith('file')) {
      return Image.file(File(url), fit: fit, width: width, height: height,
          errorBuilder: (c, o, s) {
        return _buildErrorWidget(context, msg: "File error path=$url");
      });
    } else if (url.startsWith('data:img/jpg;base64') ||
        url.startsWith('data:image/png;base64')) {
      try {
        return Image.memory(Base64Decoder().convert(url.split(',')[1]),
            fit: fit, width: width, height: height);
      } catch (e) {
        return _buildErrorWidget(context, msg: "Base64Decoder error");
      }
    } else {
      return _buildErrorWidget(context, msg: "other error");
    }
  }

  Widget _buildErrorWidget(BuildContext context, {String msg = ''}) {
    debugPrint("ImageWidget msg:$msg $this)}");
    if (errorBuild != null) return errorBuild!(context);
    return ImageWidgetComponent.instance.errorWidgetBuilder?.call(this) ??
        Container(
            decoration: BoxDecoration(
              image: DecorationImage(
                image: NetworkImage(
                    "https://www.google.com/images/branding/googlelogo/2x/googlelogo_color_92x30dp.png"),
                fit: BoxFit.cover,
              ),
            ),
            child: Text(
              msg,
              style: TextStyle(color: Colors.white60),
            ),
            width: width,
            height: height);
  }

  @override
  Widget _childTransformBuilder(Widget superCurrent) {
    // TODO: implement _childTransformBuilder
    return super._childTransformBuilder(superCurrent);
  }
}
